/*
=================================================
|                                               |
|      03 UMAP 1D                               |
|                                               |
=================================================

In this script you can perform the smae kind of UMAP 1-dimensional dimensionality reduction
used by Moore for his piece.
*/

// 0. OPTIONS AND LOAD BUFFER =========================================================================================
s.options.outDevice_("MacBook Pro Speakers");
s.options.sampleRate_(44100);
s.options.outDevice_("External Headphones");
s.options.memSize = 8192 * 8;
s.killAll;

~src = Buffer.read(s,FluidFilesPath("Nicol-LoopE-M.wav"));
// ~src = Buffer.read(s,FluidFilesPath("Green-Box641.wav"));
// ~src = Buffer.read(s,FluidFilesPath("Harker-DS-TenOboeMultiphonics-M.wav"));
// ~src = Buffer.read(s,FluidFilesPath("Olencki-TenTromboneLongTones-M.wav"));

// Have a listen:
~src.play;

// 1. ANALYSIS ========================================================================================================
(
var current_frame = 0;
var featuresBuf = Buffer(s);
var stats = Buffer(s);
var flat = Buffer(s);

// We use a regular hop size for the "slicing" of our source:
//~hopSize = 0.1 * ~src.sampleRate;
~hopSize = 512;
~windowSize = ~hopSize * 2;

~ds = FluidDataSet(s);
~counter = 0;

fork{
	while({
		(current_frame + ~windowSize) < ~src.numFrames;
	},{
		// Use MFCC:
		FluidBufMFCC.processBlocking(
			s,
			~src,
			current_frame,
			~hopSize * 2,
			windowSize:~windowSize,
			features:featuresBuf,
			startCoeff:1,
			numCoeffs:39

		);
		// Or any of the other FluCoMa Descriptors:
		// FluidBufChroma.processBlocking(s,~src,current_frame,~hopSize,features:featuresBuf);
		// FluidBufPitch.processBlocking(s,~src,current_frame,~hopSize,features:featuresBuf,unit:1,windowSize:2048);
		// FluidBufSpectralShape.processBlocking(s,~src,current_frame,~hopSize,features:featuresBuf,unit:1);
		// FluidBufStats.processBlocking(s,featuresBuf,startFrame:1,numFrames:1,stats:stats);

		FluidBufFlatten.processBlocking(
			s,
			featuresBuf,
			startFrame:1,
			numFrames:1,
			destination:flat);

		~ds.addPoint(current_frame.asInteger,flat);

		if((~counter % 100) == 99) {
			s.sync;
			~counter.postln;
			featuresBuf.postln;
			flat.postln;
			"Current frame: % / %\t%".format(current_frame,~src.numFrames,current_frame/~src.numFrames).postln;
		};

		current_frame = current_frame + ~hopSize;
		~counter = ~counter + 1;
	})
}
)

// Have a look at the created dataset:
~ds.print;

// Scale if our dimensions are in different ranges:
FluidStandardize(s).fitTransform(~ds,~ds);
~ds.print;

// 2. UMAP REDUCTION ==================================================================================================
(
// The number of dimensions given is 1:
~ds_umap = FluidDataSet(s);
FluidUMAP(s,1).fitTransform(
	~ds,~ds_umap,
	{"UMAP complete!".postln;});
)

// 3. SORT UMAP DATASET ===============================================================================================
(
// Now we reorder the dataset in ascending order according to the value each slice was given from
// the dimensionality reduction:
~ds_umap.dump({
	arg dict;
	var points = dict["data"].keys.collect{
		arg k;
		[k,dict["data"][k][0]]
	}.asArray;
	var sorted_points = points.sort({arg a, b; a[1] < b[1]}).collect{arg pt; pt[0].interpret};

	~sorted_start_frames = Buffer.loadCollection(s,sorted_points);
	sorted_points.postln;
})
)

~sorted_start_frames.plot

// 4. RECOMBINE =======================================================================================================
(
// Finally we can stitch the fragments back together in this ascending order:
fork({
	var overlap = 2;
	var hopDur = ~hopSize / ~src.sampleRate;
	var trigFreq = hopDur.reciprocal;
	var win, uv, startFrame = 0;

	Window.closeAll;

	win = Window(bounds:Rect(0,0,1600,400));
	FluidWaveform(~src,parent:win,bounds:Rect(0,0,win.bounds.width,win.bounds.height));
	1.wait;
	uv = UserView(win,Rect(0,0,win.bounds.width,win.bounds.height))
	.drawFunc_{
		var x = startFrame.linlin(0,~src.numFrames,0,win.bounds.width);
		Pen.color_(Color.red);
		Pen.addRect(Rect(x,0,2,win.bounds.height));
		Pen.fill;
	};

	{
		var trig = Impulse.ar(trigFreq);
		var index = (PulseCount.ar(trig) - 1);
		var startFrame_ = Index.kr(~sorted_start_frames,index);
		var centerFrame = startFrame_ + ~hopSize;
		var centerSeconds = centerFrame / ~src.sampleRate;
		var dur = ~windowSize / ~src.sampleRate;
		FreeSelf.kr(index > BufFrames.ir(~sorted_start_frames));
		SendReply.ar(trig,"/startFrame",startFrame_);
		TGrains.ar(1,trig,~src,BufRateScale.ir(~src),centerSeconds,dur);
	}.play;

	OSCdef(\startFrame,{
		arg msg;
		startFrame = msg[3];
		{win.refresh}.defer;
	},"/startFrame");

	win.front;
},AppClock);
)